@@ -23,10 +23,20 @@
 from typing import Dict
 from typing import List
 from typing import Tuple
+from typing import NamedTuple
 from typing import Iterable
 from typing import Any
 import sys
 import ctypes
+import logging
+from gi import require_version # type: ignore
+# pylint: disable=wrong-import-position
+require_version('IBus', '1.0')
+from gi.repository import IBus # type: ignore
+# pylint: enable=wrong-import-position
+import itb_util
+
+LOGGER = logging.getLogger('ibus-typing-booster')
 
 # pylint: disable=invalid-name
 # pylint: disable=too-few-public-methods
@@ -907,112 +917,47 @@ def convert_digits_to_ascii(text: str) -> str:
     '''
     return text.translate(DIGIT_TRANS_TABLE)
 
+class TransliterationParts(NamedTuple):
+    '''
+    A named tuple containing the parts of a transliteration
+
+    committed: str        The part of the transliteration which cannot be
+                          changed anymore by adding more input, could be
+                          committed already if desired.
+    committed_index: int  The index up to which the msymbol_list input
+                          was “used up” to create the “committed” text.
+    preedit:              The transliteration of the remaining input,
+                          may still change by adding more input.
+    cursor_pos: int       The cursor position in the preedit.
+                          Counted in codepoints, not glyphs.
+                          Usually this is at the end of the preedit
+                          but an input method may move the cursor
+                          within the preedit!
+                          (I think only ja-anthy.mim actually uses this)
+    status: str           May change for some input methods to
+                          indicate a state.
+                          For example in case of ja-anthy.mim,
+                          this is 'aあ' before Henkan and changes
+                          to '漢' in Henkan mode.
+    candidates: List[str] May contain a list of candidates if the
+                          input method can produce multiple candidates.
+    '''
+    committed: str = ''
+    committed_index: int = 0
+    preedit: str = ''
+    cursor_pos: int = 0
+    status: str = ''
+    candidates: List[str] = []
+
 class Transliterator:
+    # pylint: disable=line-too-long
     '''A class for transliterators using libm17n
 
     If initializing the transliterator fails, for example because a
     non-existing input method was given as the argument, a ValueError
-    is raised:
-
-    Examples:
-
-    Russian transliteration:
-
-    >>> trans = Transliterator('ru-translit')
-    >>> trans.transliterate(list('y'))
-    'ы'
-    >>> trans.transliterate(list('yo'))
-    'ё'
-    >>> trans.transliterate(list('yo y'))
-    'ё ы'
-
-    Marathi transliteration:
-
-    >>> trans = Transliterator('mr-itrans')
-    >>> trans.transliterate(list('praviN'))
-    'प्रविण्'
-    >>> trans.transliterate(list('namaste'))
-    'नमस्ते'
-
-    Hindi transliteration:
-
-    >>> trans = Transliterator('hi-itrans')
-    >>> trans.transliterate(list('namaste'))
-    'नमस्ते'
-
-    >>> trans.transliterate_parts(list('n'))
-    ('', 'न्')
-
-    >>> trans.transliterate_parts(list('n '))
-    ('न ', '')
-
-    >>> trans.transliterate_parts(list('na'))
-    ('', 'न')
-
-    >>> trans.transliterate_parts(list('nam'))
-    ('न', 'म्')
-
-    >>> trans.transliterate_parts(list('nama'))
-    ('न', 'म')
-
-    >>> trans.transliterate_parts(list('namas'))
-    ('नम', 'स्')
-
-    >>> trans.transliterate_parts(list('namast'))
-    ('नम', 'स्त्')
-
-    >>> trans.transliterate_parts(list('namaste'))
-    ('नम', 'स्ते')
-
-    >>> trans.transliterate_parts(list('namaste '))
-    ('नमस्ते ', '')
-
-    >>> trans.transliterate(list('. '))
-    '। '
-
-    Hindi-Inscript2 uses the AltGr key a lot, 'G-4' is
-    the MSymbol name for AltGr-4 and it transliterates
-    to something different than just '4':
-
-    >>> trans = Transliterator('hi-inscript2')
-    >>> trans.transliterate(['4', 'G-4'])
-    '४₹'
-
-    >>> trans = Transliterator('hi-inscript2')
-    >>> trans.transliterate(['G-p'])
-    'ज़'
-
-    AltGr-3 ('G-3') is not used though in Hindi-Inscript2.
-    Therefore, 'G-3' transliterates just as 'G-3':
-
-    >>> trans = Transliterator('hi-inscript2')
-    >>> trans.transliterate(['3', 'G-3'])
-    '३G-3'
-
-    In mr-inscript2, 'G-1' transliterates to U+200D ZERO WIDTH JOINER
-    ('\xe2\x80\x8d' in UTF-8 encoding):
-
-    >>> trans = Transliterator('mr-inscript2')
-    >>> trans.transliterate(['j', 'd', 'G-1', '/']).encode('utf-8')
-    b'\xe0\xa4\xb0\xe0\xa5\x8d\xe2\x80\x8d\xe0\xa4\xaf'
-
-    >>> trans = Transliterator('t-latn-post')
-    >>> trans.transliterate(list('gru"n'))
-    'grün'
-
-    >>> trans = Transliterator('NoIME')
-    >>> trans.transliterate(['a', 'b', 'c', 'C-c', 'G-4', 'C-α', 'G-α'])
-    'abcC-cG-4C-αG-α'
-
-    >>> trans = Transliterator('ko-romaja')
-    >>> trans.transliterate(list('annyeonghaseyo'))
-    '안녕하세요'
-
-    >>> trans = Transliterator('si-wijesekara')
-    >>> trans.transliterate(list('vksIal kjSka '))
-    'ඩනිෂ්ක නවීන් '
-
+    is raised.
     '''
+    # pylint: enable=line-too-long
     def __init__(self, ime: str) -> None:
         '''Initialize the input method to use for the transliteration
 
@@ -1045,7 +990,10 @@ class Transliterator:
         except ValueError as error: # NULL pointer access
             raise ValueError('minput_create_ic() failed') from error
 
-    def transliterate_parts(self, msymbol_list: Iterable[str], ascii_digits: bool = False) -> Tuple[str, str]:
+    def transliterate_parts(
+            self,
+            msymbol_list: Iterable[str],
+            ascii_digits: bool = False) -> TransliterationParts:
         '''Transliterate a list of Msymbol names
 
         :param msymbol_list: A list of strings which are interpreted
@@ -1055,16 +1003,90 @@ class Transliterator:
                              Msymbols is just joined to a single string.
         :param ascii_digits: If true, convert language specific digits
                              to ASCII digits
-        :return: The transliteration in two parts: (committed, preedit)
+        :return: The transliteration in several parts
+
+        Examples:
+
+        >>> trans = Transliterator('hi-itrans')
+        >>> trans.transliterate_parts(list('n'))
+        TransliterationParts(committed='', committed_index=0, preedit='न्', cursor_pos=2, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('n '))
+        TransliterationParts(committed='न ', committed_index=2, preedit='', cursor_pos=0, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('na'))
+        TransliterationParts(committed='', committed_index=0, preedit='न', cursor_pos=1, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('nam'))
+        TransliterationParts(committed='न', committed_index=2, preedit='म्', cursor_pos=2, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('nama'))
+        TransliterationParts(committed='न', committed_index=2, preedit='म', cursor_pos=1, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('namas'))
+        TransliterationParts(committed='नम', committed_index=4, preedit='स्', cursor_pos=2, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('namast'))
+        TransliterationParts(committed='नम', committed_index=4, preedit='स्त्', cursor_pos=4, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('namaste'))
+        TransliterationParts(committed='नम', committed_index=4, preedit='स्ते', cursor_pos=4, status='क', candidates=[])
+        >>> trans.transliterate_parts(list('namaste '))
+        TransliterationParts(committed='नमस्ते ', committed_index=8, preedit='', cursor_pos=0, status='क', candidates=[])
+
+        >>> trans = Transliterator('t-latn-post')
+        >>> trans.transliterate_parts(list('u'))
+        TransliterationParts(committed='', committed_index=0, preedit='u', cursor_pos=1, status='Latin-post', candidates=[])
+        >>> trans.transliterate_parts(list('u"'))
+        TransliterationParts(committed='', committed_index=0, preedit='ü', cursor_pos=1, status='Latin-post', candidates=[])
+        >>> trans.transliterate_parts(list('u""'))
+        TransliterationParts(committed='u"', committed_index=3, preedit='', cursor_pos=0, status='Latin-post', candidates=[])
+        >>> trans.transliterate_parts(list('u"u'))
+        TransliterationParts(committed='ü', committed_index=2, preedit='u', cursor_pos=1, status='Latin-post', candidates=[])
+        >>> trans.transliterate_parts(list('üu"u'))
+        TransliterationParts(committed='üü', committed_index=3, preedit='u', cursor_pos=1, status='Latin-post', candidates=[])
+
+        >>> trans = Transliterator('t-rfc1345')
+        >>> trans.transliterate_parts(list('&'))
+        TransliterationParts(committed='', committed_index=0, preedit='&', cursor_pos=1, status='RFC1345', candidates=[])
+        >>> trans.transliterate_parts(list('&C'))
+        TransliterationParts(committed='', committed_index=0, preedit='&C', cursor_pos=2, status='RFC1345', candidates=[])
+        >>> trans.transliterate_parts(list('&Co'))
+        TransliterationParts(committed='©', committed_index=3, preedit='', cursor_pos=0, status='RFC1345', candidates=[])
+        >>> trans.transliterate_parts(list('&f'))
+        TransliterationParts(committed='', committed_index=0, preedit='&f', cursor_pos=2, status='RFC1345', candidates=[])
+        >>> trans.transliterate_parts(list('&ff'))
+        TransliterationParts(committed='', committed_index=0, preedit='ﬀ', cursor_pos=1, status='RFC1345', candidates=[])
+        >>> trans.transliterate_parts(list('&ffi'))
+        TransliterationParts(committed='ﬃ', committed_index=4, preedit='', cursor_pos=0, status='RFC1345', candidates=[])
+
+        >>> trans = Transliterator('t-lsymbol')
+        >>> trans.transliterate_parts(list('/:)'))
+        TransliterationParts(committed='', committed_index=0, preedit='☺️', cursor_pos=2, status='lsymbol', candidates=['☺️', '😃', '😅', '😆', '😉', '😇', '😂', '😏', '😛', '😜', '😝', '😋', '😉', '💏', '💋', '😍', '😘', '😚', '😽', '😻'])
+        >>> trans.transliterate_parts(list('a'))
+        TransliterationParts(committed='a', committed_index=1, preedit='', cursor_pos=0, status='lsymbol', candidates=[])
+        >>> trans.transliterate_parts(list('a/'))
+        TransliterationParts(committed='a', committed_index=1, preedit='/', cursor_pos=1, status='lsymbol', candidates=[])
+        >>> trans.transliterate_parts(list('a/:'))
+        TransliterationParts(committed='a', committed_index=1, preedit='/:', cursor_pos=2, status='lsymbol', candidates=[])
+        >>> trans.transliterate_parts(list('a/:('))
+        TransliterationParts(committed='a', committed_index=1, preedit='😢', cursor_pos=1, status='lsymbol', candidates=['😢', '😩', '😡', '😭', '😪', '🙈', '🙊', '🙉'])
+        >>> trans.transliterate_parts(list('a/:(b'))
+        TransliterationParts(committed='a😢b', committed_index=5, preedit='', cursor_pos=0, status='lsymbol', candidates=[])
+
+        For a test transliterating parts using 'ja-anthy' see 'tests/test_m17n_translit.py'.
         '''
         if not isinstance(msymbol_list, list):
             raise ValueError('Argument of transliterate() must be a list.')
         if self._dummy:
-            return (''.join(msymbol_list), '')
+            return TransliterationParts(committed=''.join(msymbol_list),
+                                       committed_index=len(msymbol_list))
         libm17n__minput_reset_ic(self._ic) # type: ignore
         committed = ''
+        committed_index = 0
         preedit = ''
-        for symbol in msymbol_list:
+        candidates: List[str] = []
+        for index, symbol in enumerate(msymbol_list):
+            if len(symbol) == 1 and not itb_util.is_ascii(symbol):
+                symbol = IBus.keyval_name(IBus.unicode_to_keyval(symbol))
+            elif (len(symbol) == 3 and symbol[1] == '-'
+                and symbol[0] in ('G', 'C', 'A')
+                and not itb_util.is_ascii(symbol[2])):
+                symbol = symbol[:2] + IBus.keyval_name(
+                    IBus.unicode_to_keyval(symbol[2]))
             _symbol = libm17n__msymbol(symbol.encode('utf-8')) # type: ignore
             retval = libm17n__minput_filter( # type: ignore
                 self._ic, _symbol, ctypes.c_void_p(None))
@@ -1074,8 +1096,10 @@ class Transliterator:
                     self._ic, _symbol, ctypes.c_void_p(None), _mt)
                 if libm17n__mtext_len(_mt) > 0: # type: ignore
                     committed += mtext_to_string(_mt)
+                    committed_index = index
                 if retval:
-                    committed += symbol
+                    committed += msymbol_list[index]
+                    committed_index = index + 1
         try:
             if (self._ic.contents.preedit_changed
                 and
@@ -1085,6 +1109,42 @@ class Transliterator:
         except Exception as error: # pylint: disable=broad-except
             # This should never happen:
             raise ValueError('Problem accessing preedit') from error
+        plist = self._ic.contents.candidate_list
+        while bool(plist):  # NULL pointers have a False boolean value
+            key = libm17n__mplist_key(plist) # type: ignore
+            if not bool(key):
+                break
+            key_name = libm17n__msymbol_name(key.contents) # type: ignore
+            if key_name == b'mtext':
+                characters = mtext_to_string(
+                    ctypes.cast(libm17n__mplist_value(plist), # type: ignore
+                        ctypes.POINTER(libm17n__MText)))
+                candidates += list(characters)
+            elif key_name == b'plist':
+                candidate_plist = ctypes.cast(
+                    libm17n__mplist_value(plist), # type: ignore
+                    ctypes.POINTER(libm17n__MPlist))
+                while True:
+                    candidate_plist_key = libm17n__mplist_key( # type: ignore
+                        candidate_plist)
+                    if not bool(candidate_plist_key):
+                        break
+                    candidate_plist_key_name = libm17n__msymbol_name( # type: ignore
+                        candidate_plist_key.contents)
+                    if candidate_plist_key_name != b'mtext':
+                        break
+                    candidate = mtext_to_string(
+                        ctypes.cast(
+                            libm17n__mplist_value(candidate_plist), # type: ignore
+                            ctypes.POINTER(libm17n__MText)))
+                    candidates.append(candidate)
+                    candidate_plist = libm17n__mplist_next( # type: ignore
+                        candidate_plist)
+            else:
+                break
+            plist = libm17n__mplist_next(plist) # type: ignore
+        cursor_pos = self._ic.contents.cursor_pos
+        status = mtext_to_string(self._ic.contents.status)
         # From the m17n-lib documentation:
         #
         # The minput_reset_ic () function resets input context $IC by
@@ -1109,10 +1169,22 @@ class Transliterator:
         _symbol = libm17n__msymbol(b'nil') # type: ignore
         _retval = libm17n__minput_filter( # type: ignore
             self._ic, _symbol, ctypes.c_void_p(None))
+        if committed and not preedit:
+            committed_index = len(msymbol_list)
         if not ascii_digits:
-            return (committed, preedit)
-        return (convert_digits_to_ascii(committed),
-                convert_digits_to_ascii(preedit))
+            return TransliterationParts(committed=committed,
+                                       committed_index=committed_index,
+                                       preedit=preedit,
+                                       cursor_pos=cursor_pos,
+                                       status=status,
+                                       candidates=candidates)
+        return TransliterationParts(
+            committed=convert_digits_to_ascii(committed),
+            committed_index=committed_index,
+            preedit=convert_digits_to_ascii(preedit),
+            cursor_pos=cursor_pos,
+            status=status,
+            candidates=candidates)
 
     def transliterate(self, msymbol_list: Iterable[str], ascii_digits: bool = False) -> str:
         '''Transliterate a list of Msymbol names
@@ -1125,9 +1197,80 @@ class Transliterator:
         :param ascii_digits: If true, convert language specific digits
                              to ASCII digits
         :return: The transliteration in one string
+
+        Examples:
+
+        Russian transliteration:
+
+        >>> trans = Transliterator('ru-translit')
+        >>> trans.transliterate(list('y'))
+        'ы'
+        >>> trans.transliterate(list('yo'))
+        'ё'
+        >>> trans.transliterate(list('yo y'))
+        'ё ы'
+
+        Marathi transliteration:
+
+        >>> trans = Transliterator('mr-itrans')
+        >>> trans.transliterate(list('praviN'))
+        'प्रविण्'
+        >>> trans.transliterate(list('namaste'))
+        'नमस्ते'
+
+        Hindi transliteration:
+
+        >>> trans = Transliterator('hi-itrans')
+        >>> trans.transliterate(list('namaste'))
+        'नमस्ते'
+        >>> trans.transliterate(list('. '))
+        '। '
+
+        Hindi-Inscript2 uses the AltGr key a lot, 'G-4' is
+        the MSymbol name for AltGr-4 and it transliterates
+        to something different than just '4':
+
+        >>> trans = Transliterator('hi-inscript2')
+        >>> trans.transliterate(['4', 'G-4'])
+        '४₹'
+
+        >>> trans = Transliterator('hi-inscript2')
+        >>> trans.transliterate(['G-p'])
+        'ज़'
+
+        AltGr-3 ('G-3') is not used though in Hindi-Inscript2.
+        Therefore, 'G-3' transliterates just as 'G-3':
+
+        >>> trans = Transliterator('hi-inscript2')
+        >>> trans.transliterate(['3', 'G-3'])
+        '३G-3'
+
+        In mr-inscript2, 'G-1' transliterates to U+200D ZERO WIDTH JOINER
+        ('\xe2\x80\x8d' in UTF-8 encoding):
+
+        >>> trans = Transliterator('mr-inscript2')
+        >>> trans.transliterate(['j', 'd', 'G-1', '/']).encode('utf-8')
+        b'\xe0\xa4\xb0\xe0\xa5\x8d\xe2\x80\x8d\xe0\xa4\xaf'
+
+        >>> trans = Transliterator('t-latn-post')
+        >>> trans.transliterate(list('gru"n'))
+        'grün'
+
+        >>> trans = Transliterator('NoIME')
+        >>> trans.transliterate(['a', 'b', 'c', 'C-c', 'G-4', 'C-α', 'G-α'])
+        'abcC-cG-4C-αG-α'
+
+        >>> trans = Transliterator('ko-romaja')
+        >>> trans.transliterate(list('annyeonghaseyo'))
+        '안녕하세요'
+
+        >>> trans = Transliterator('si-wijesekara')
+        >>> trans.transliterate(list('vksIal kjSka '))
+        'ඩනිෂ්ක නවීන් '
         '''
-        (committed, preedit) = self.transliterate_parts(msymbol_list, ascii_digits)
-        return committed + preedit
+        transliteration_parts = self.transliterate_parts(
+            msymbol_list, ascii_digits)
+        return transliteration_parts.committed + transliteration_parts.preedit
 
     def get_variables(self) -> List[Tuple[str, str, str]]:
         # pylint: disable=line-too-long
@@ -1329,6 +1472,9 @@ class Transliterator:
         return
 
 if __name__ == "__main__":
+    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
+    LOGGER.setLevel(logging.DEBUG)
+    LOGGER.addHandler(LOG_HANDLER)
     import doctest
-    (FAILED, _ATTEMPTED, *_EXTRAS) = doctest.testmod()
+    (FAILED, _ATTEMPTED) = doctest.testmod()
     sys.exit(FAILED)
